from datetime import date


class DateRange:
    def __init__(
        self,
        start: date | None = None,
        end: date | None = None,
    ):
        self.start = start
        self.end = end

    @staticmethod
    def from_str(string: str):
        try:
            start_date_str, end_date_str = string.split("..")
            start_date = (
                date.fromisoformat(start_date_str) if start_date_str != "" else None
            )
            end_date = date.fromisoformat(end_date_str) if end_date_str != "" else None
            return DateRange(start_date, end_date)
        except ValueError:
            raise ValueError(
                "Invalid date range string format. Expected 'YYYY-MM-DD..YYYY-MM-DD'."
            )

    def __str__(self):
        start_str = self.start if self.start is not None else ""
        end_str = self.end if self.end is not None else ""
        return f"{start_str}..{end_str}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return (self.start, self.end) == (other.start, other.end)

        return False


def convert_date_range_to_str_for_search_query(date_range: DateRange) -> str | None:
    if date_range.start is not None and date_range.end is not None:
        return str(date_range)
    elif date_range.start is not None and date_range.end is None:
        return f">={str(date_range.start)}"
    elif date_range.start is None and date_range.end is not None:
        return f"<={str(date_range.end)}"
    else:
        return None



from datetime import datetime
import json
from typing import Optional


class History:
    def __init__(
        self,
        first_committed_at: datetime,
        merged_at: datetime,
        deployed_at: Optional[datetime],
    ):
        self.first_committed_at = first_committed_at
        self.merged_at = merged_at
        self.deployed_at = deployed_at

    @staticmethod
    def keys() -> list[str]:
        return [
            "first_committed_at",
            "merged_at",
            "deployed_at",
        ]

    def values(self) -> list:
        return [self.first_committed_at, self.merged_at, self.deployed_at]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, History):
            return NotImplemented
        return (
            self.first_committed_at == other.first_committed_at
            and self.merged_at == other.merged_at
            and self.deployed_at == other.deployed_at
        )


class RepositoryHistory:
    def __init__(
        self,
        first_committed_at: datetime,
        merged_at: datetime,
        merge_commit_sha: str,
        check_runs: list,
    ):
        self.first_committed_at = first_committed_at
        self.merged_at = merged_at
        self.merge_commit_sha = merge_commit_sha
        self.check_runs = check_runs

    def __get_target_check_run(self, name: str | None = None) -> dict | None:
        if name is None:
            return self.check_runs[0]

        try:
            return next(cr for cr in self.check_runs if cr["name"] == name)
        except StopIteration:
            return None

    def get_workflow_id(self, name: str | None = None) -> str | None:
        if len(self.check_runs) == 0:
            return None

        check_run = self.__get_target_check_run(name)

        if check_run is None:
            return None

        if check_run["app"]["slug"] != "circleci-checks":
            return None

        decoded_external_id = json.loads(check_run["external_id"])

        return decoded_external_id["workflow-id"]


def convert_repository_history_to_workflow_ids(
    repository_histories: list[RepositoryHistory], workflow_name: str | None = None
) -> list[str]:
    raw_workflow_ids = [
        history.get_workflow_id(workflow_name) for history in repository_histories
    ]
    return [x for x in raw_workflow_ids if x is not None]


class DeploymentHistory:
    def __init__(
        self,
        id: str,  # workflow id
        deployed_at: datetime,
    ):
        self.id = id
        self.deployed_at = deployed_at


def create_histories(
    repository_histories: list[RepositoryHistory],
    deployment_histories: list[DeploymentHistory],
    workflow_name: str | None = None,
) -> list[History]:
    idToDeploymentHistoryMap: dict[str, DeploymentHistory] = {}
    for dh in deployment_histories:
        idToDeploymentHistoryMap[dh.id] = dh

    histories = []
    for rh in repository_histories:
        workflow_id = rh.get_workflow_id(workflow_name)
        matched_dh = (
            idToDeploymentHistoryMap.get(workflow_id)
            if workflow_id is not None
            else None
        )
        histories.append(
            History(
                rh.first_committed_at,
                rh.merged_at,
                matched_dh.deployed_at if matched_dh is not None else None,
            )
        )

    return histories



from datetime import date
from depwatch.date_utils import DateRange, convert_date_range_to_str_for_search_query


class TestDateUtils:
    def test_date_range_from_str(self):
        datasets = [
            "2023-01-01..2023-03-31",
            "2023-01-01..",
            "..2023-03-31",
        ]

        for string in datasets:
            date_range = DateRange.from_str(string)

            assert str(date_range) == string

    def test_date_range_from_str_when_invalid_string_is_provided(self):
        try:
            DateRange.from_str("2023-01-01")
            assert False
        except ValueError as e:
            assert (
                str(e)
                == "Invalid date range string format. Expected 'YYYY-MM-DD..YYYY-MM-DD'."
            )

    def test_date_convert_date_range_to_str_for_search_query(self):
        datasets = [
            [
                date.fromisoformat("2023-01-01"),
                date.fromisoformat("2023-03-31"),
                "2023-01-01..2023-03-31",
            ],
            [date.fromisoformat("2023-01-01"), None, ">=2023-01-01"],
            [None, date.fromisoformat("2023-03-31"), "<=2023-03-31"],
            [None, None, None],
        ]

        for start, end, string in datasets:
            date_range = DateRange(start, end)

            assert convert_date_range_to_str_for_search_query(date_range) == string



from datetime import datetime
from depwatch.history import (
    DeploymentHistory,
    History,
    RepositoryHistory,
    convert_repository_history_to_workflow_ids,
    create_histories,
)


class TestHistory:
    def test_repository_history_get_workflow_id(self) -> None:
        rh = RepositoryHistory(
            datetime.fromisoformat("2023-04-26T20:00:00+00:00"),
            datetime.fromisoformat("2023-04-26T22:00:00+00:00"),
            "b30000653abd86ff77b2a3e59b427896f4b16ccf",
            [
                {
                    "name": "awesome_workflow",
                    "external_id": '{"workflow-id":"72d1d896-cfbe-48e2-a023-51edb6438b9f","actor-id":"5ead47d7-6407-4264-a150-7f68b8c98631"}',
                    "app": {
                        "slug": "circleci-checks",
                    },
                },
            ],
        )

        assert rh.get_workflow_id() == "72d1d896-cfbe-48e2-a023-51edb6438b9f"

    def test_repository_history_get_workflow_id_when_check_runs_is_empty(
        self,
    ) -> None:
        rh = RepositoryHistory(
            datetime.fromisoformat("2023-04-26T20:00:00+00:00"),
            datetime.fromisoformat("2023-04-26T22:00:00+00:00"),
            "b30000653abd86ff77b2a3e59b427896f4b16ccf",
            [],
        )

        assert rh.get_workflow_id() == None

    def test_repository_history_get_workflow_id_when_the_check_run_app_slug_is_not_supported(
        self,
    ) -> None:
        rh = RepositoryHistory(
            datetime.fromisoformat("2023-04-26T20:00:00+00:00"),
            datetime.fromisoformat("2023-04-26T22:00:00+00:00"),
            "b30000653abd86ff77b2a3e59b427896f4b16ccf",
            [
                {
                    "name": "awesome_workflow",
                    "external_id": '{"foo":"bar"}',
                    "app": {
                        "slug": "custom-checks",
                    },
                },
            ],
        )

        assert rh.get_workflow_id() == None

    def test_repository_history_get_workflow_id_when_workflow_name_is_specified(
        self,
    ) -> None:
        rh = RepositoryHistory(
            datetime.fromisoformat("2023-04-26T20:00:00+00:00"),
            datetime.fromisoformat("2023-04-26T22:00:00+00:00"),
            "b30000653abd86ff77b2a3e59b427896f4b16ccf",
            [
                {
                    "name": "not_target",
                    "external_id": '{"workflow-id":"72d1d896-cfbe-48e2-a023-51edb6438b9f","actor-id":"5ead47d7-6407-4264-a150-7f68b8c98631"}',
                    "app": {
                        "slug": "circleci-checks",
                    },
                },
                {
                    "name": "target",
                    "external_id": '{"workflow-id":"0ba8c279-3f07-40a2-bf86-5df5e60f3c74","actor-id":"d5f1a47d-867a-46c6-8bbc-0b38f908d4f7"}',
                    "app": {
                        "slug": "circleci-checks",
                    },
                },
                {
                    "name": "not_target",
                    "external_id": '{"workflow-id":"c7b22720-23e0-49f8-9b43-47b4cf457a8f","actor-id":"84f7de2f-ca21-4609-83ad-bbc148298b82"}',
                    "app": {
                        "slug": "circleci-checks",
                    },
                },
            ],
        )

        assert rh.get_workflow_id("target") == "0ba8c279-3f07-40a2-bf86-5df5e60f3c74"

    def test_repository_history_get_workflow_id_when_workflow_name_is_specified_but_does_not_exist(
        self,
    ) -> None:
        rh = RepositoryHistory(
            datetime.fromisoformat("2023-04-26T20:00:00+00:00"),
            datetime.fromisoformat("2023-04-26T22:00:00+00:00"),
            "b30000653abd86ff77b2a3e59b427896f4b16ccf",
            [
                {
                    "name": "not_target",
                    "external_id": '{"workflow-id":"72d1d896-cfbe-48e2-a023-51edb6438b9f","actor-id":"5ead47d7-6407-4264-a150-7f68b8c98631"}',
                    "app": {
                        "slug": "circleci-checks",
                    },
                },
            ],
        )

        assert rh.get_workflow_id("target") == None

    def test_convert_repository_history_to_workflow_ids(self) -> None:
        repository_histories = [
            RepositoryHistory(
                datetime.fromisoformat("2023-04-26T20:00:00+00:00"),
                datetime.fromisoformat("2023-04-26T22:00:00+00:00"),
                "b30000653abd86ff77b2a3e59b427896f4b16ccf",
                [
                    {
                        "name": "awesome_workflow",
                        "external_id": '{"workflow-id":"72d1d896-cfbe-48e2-a023-51edb6438b9f","actor-id":"5ead47d7-6407-4264-a150-7f68b8c98631"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                ],
            ),
            RepositoryHistory(
                datetime.fromisoformat("2023-04-25T20:00:00+00:00"),
                datetime.fromisoformat("2023-04-25T22:00:00+00:00"),
                "cebc0584d4ebeb2da4ed488be5d02c9ca2c7efeb",
                [
                    {
                        "name": "awesome_workflow",
                        "external_id": '{"workflow-id":"da5718a4-d796-4414-a6f7-bb3b26c01bf8","actor-id":"33a61365-e7ef-453c-a105-4aa8d342ca8e"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                ],
            ),
            RepositoryHistory(
                datetime.fromisoformat("2023-04-24T20:00:00+00:00"),
                datetime.fromisoformat("2023-04-24T22:00:00+00:00"),
                "a6a8e9e33c0854dc742f03e0c48977cad1cec72f",
                [
                    {
                        "name": "awesome_workflow",
                        "external_id": '{"workflow-id":"20219ed8-6bce-4760-9b7c-b50d2eb34099","actor-id":"18b815e9-4438-4fba-b65c-77b224ecc116"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                ],
            ),
        ]

        pipeline_list = convert_repository_history_to_workflow_ids(repository_histories)

        assert pipeline_list == [
            "72d1d896-cfbe-48e2-a023-51edb6438b9f",
            "da5718a4-d796-4414-a6f7-bb3b26c01bf8",
            "20219ed8-6bce-4760-9b7c-b50d2eb34099",
        ]

    def test_convert_repository_history_to_workflow_ids_when_workflow_name_is_specified(
        self,
    ) -> None:
        repository_histories = [
            RepositoryHistory(
                datetime.fromisoformat("2023-04-26T20:00:00+00:00"),
                datetime.fromisoformat("2023-04-26T22:00:00+00:00"),
                "b30000653abd86ff77b2a3e59b427896f4b16ccf",
                [
                    {
                        "name": "target",
                        "external_id": '{"workflow-id":"72d1d896-cfbe-48e2-a023-51edb6438b9f","actor-id":"5ead47d7-6407-4264-a150-7f68b8c98631"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                    {
                        "name": "not_target",
                        "external_id": '{"workflow-id":"0ba8c279-3f07-40a2-bf86-5df5e60f3c74","actor-id":"d5f1a47d-867a-46c6-8bbc-0b38f908d4f7"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                ],
            ),
            RepositoryHistory(
                datetime.fromisoformat("2023-04-25T20:00:00+00:00"),
                datetime.fromisoformat("2023-04-25T22:00:00+00:00"),
                "cebc0584d4ebeb2da4ed488be5d02c9ca2c7efeb",
                [
                    {
                        "name": "not_target",
                        "external_id": '{"workflow-id":"da5718a4-d796-4414-a6f7-bb3b26c01bf8","actor-id":"33a61365-e7ef-453c-a105-4aa8d342ca8e"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                ],
            ),
            RepositoryHistory(
                datetime.fromisoformat("2023-04-24T20:00:00+00:00"),
                datetime.fromisoformat("2023-04-24T22:00:00+00:00"),
                "a6a8e9e33c0854dc742f03e0c48977cad1cec72f",
                [
                    {
                        "name": "not_target",
                        "external_id": '{"workflow-id":"c7b22720-23e0-49f8-9b43-47b4cf457a8f","actor-id":"84f7de2f-ca21-4609-83ad-bbc148298b82"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                    {
                        "name": "target",
                        "external_id": '{"workflow-id":"20219ed8-6bce-4760-9b7c-b50d2eb34099","actor-id":"18b815e9-4438-4fba-b65c-77b224ecc116"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                ],
            ),
        ]

        pipeline_list = convert_repository_history_to_workflow_ids(
            repository_histories, "target"
        )

        assert pipeline_list == [
            "72d1d896-cfbe-48e2-a023-51edb6438b9f",
            "20219ed8-6bce-4760-9b7c-b50d2eb34099",
        ]

    def test_convert_repository_history_to_workflow_ids_when_a_item_not_having_workflow_id_is_included(
        self,
    ) -> None:
        repository_histories = [
            RepositoryHistory(
                datetime.fromisoformat("2023-04-26T20:00:00+00:00"),
                datetime.fromisoformat("2023-04-26T22:00:00+00:00"),
                "b30000653abd86ff77b2a3e59b427896f4b16ccf",
                [
                    {
                        "name": "awesome_workflow",
                        "external_id": '{"workflow-id":"72d1d896-cfbe-48e2-a023-51edb6438b9f","actor-id":"5ead47d7-6407-4264-a150-7f68b8c98631"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                ],
            ),
            RepositoryHistory(
                datetime.fromisoformat("2023-04-25T20:00:00+00:00"),
                datetime.fromisoformat("2023-04-25T22:00:00+00:00"),
                "cebc0584d4ebeb2da4ed488be5d02c9ca2c7efeb",
                [],
            ),
            RepositoryHistory(
                datetime.fromisoformat("2023-04-24T20:00:00+00:00"),
                datetime.fromisoformat("2023-04-24T22:00:00+00:00"),
                "a6a8e9e33c0854dc742f03e0c48977cad1cec72f",
                [
                    {
                        "name": "awesome_workflow",
                        "external_id": '{"workflow-id":"20219ed8-6bce-4760-9b7c-b50d2eb34099","actor-id":"18b815e9-4438-4fba-b65c-77b224ecc116"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                ],
            ),
        ]

        pipeline_list = convert_repository_history_to_workflow_ids(repository_histories)

        assert pipeline_list == [
            "72d1d896-cfbe-48e2-a023-51edb6438b9f",
            "20219ed8-6bce-4760-9b7c-b50d2eb34099",
        ]

    def test_create_histories(self):
        repository_histories = [
            RepositoryHistory(
                datetime.fromisoformat("2023-02-25T20:00:00+00:00"),
                datetime.fromisoformat("2023-02-26T22:00:00+00:00"),
                "8eddb4b87a1460b29bce25448623b270465e9715",
                [
                    {
                        "name": "awesome_workflow",
                        "external_id": '{"workflow-id":"52166f56-7e17-40f5-b7dd-5e8c55d388a8","actor-id":"18b815e9-4438-4fba-b65c-77b224ecc116"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                ],
            ),
            RepositoryHistory(
                datetime.fromisoformat("2023-02-23T10:05:00+00:00"),
                datetime.fromisoformat("2023-02-23T11:15:00+00:00"),
                "12d4c5f2a123b6fb77eccb57be1be3a29438fadb",
                [
                    {
                        "name": "awesome_workflow",
                        "external_id": '{"workflow-id":"b609a9ab-92f2-46b2-b1c2-a9e9b9f3ffc7","actor-id":"58f32647-4d45-4b93-9b57-e9712de5ce08"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                ],
            ),
            RepositoryHistory(
                datetime.fromisoformat("2023-02-23T10:00:00+00:00"),
                datetime.fromisoformat("2023-02-23T11:00:00+00:00"),
                "cbd1519fd8b7daff2655346676065f844a2ef3df",
                [
                    {
                        "name": "awesome_workflow",
                        "external_id": '{"workflow-id":"c0d92df0-7d09-4774-af5d-6076bc47768e","actor-id":"6f63e7b5-d4fe-41f2-8d06-5627ca3cd61a"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                ],
            ),
            RepositoryHistory(
                datetime.fromisoformat("2023-02-22T14:00:00+00:00"),
                datetime.fromisoformat("2023-02-22T17:00:00+00:00"),
                "b25ab2d8590fc678ced72fe378c0bf2dc04da15c",
                [
                    {
                        "name": "awesome_workflow",
                        "external_id": '{"workflow-id":"71fc25f3-913f-4207-bcc2-c0dc54722932","actor-id":"6a362c69-9e05-474f-b93a-89f145e3851d"}',
                        "app": {
                            "slug": "circleci-checks",
                        },
                    },
                ],
            ),
        ]
        deployment_histories = [
            DeploymentHistory(
                "52166f56-7e17-40f5-b7dd-5e8c55d388a8",
                datetime.fromisoformat("2023-02-27T10:15:00+00:00"),
            ),
            DeploymentHistory(
                "71fc25f3-913f-4207-bcc2-c0dc54722932",
                datetime.fromisoformat("2023-02-24T09:30:00+00:00"),
            ),
            DeploymentHistory(
                "b609a9ab-92f2-46b2-b1c2-a9e9b9f3ffc7",
                datetime.fromisoformat("2023-02-23T11:30:00+00:00"),
            ),
        ]

        histories = create_histories(repository_histories, deployment_histories)

        assert histories == [
            History(
                datetime.fromisoformat("2023-02-25T20:00:00+00:00"),
                datetime.fromisoformat("2023-02-26T22:00:00+00:00"),
                datetime.fromisoformat("2023-02-27T10:15:00+00:00"),
            ),
            History(
                datetime.fromisoformat("2023-02-23T10:05:00+00:00"),
                datetime.fromisoformat("2023-02-23T11:15:00+00:00"),
                datetime.fromisoformat("2023-02-23T11:30:00+00:00"),
            ),
            History(
                datetime.fromisoformat("2023-02-23T10:00:00+00:00"),
                datetime.fromisoformat("2023-02-23T11:00:00+00:00"),
                None,
            ),
            History(
                datetime.fromisoformat("2023-02-22T14:00:00+00:00"),
                datetime.fromisoformat("2023-02-22T17:00:00+00:00"),
                datetime.fromisoformat("2023-02-24T09:30:00+00:00"),
            ),
        ]

    def test_create_histories_when_a_repository_history_with_no_workflow_id_exists(
        self,
    ):
        repository_histories = [
            RepositoryHistory(
                datetime.fromisoformat("2023-02-25T20:00:00+00:00"),
                datetime.fromisoformat("2023-02-26T22:00:00+00:00"),
                "8eddb4b87a1460b29bce25448623b270465e9715",
                [],
            ),
        ]
        deployment_histories = [
            DeploymentHistory(
                "52166f56-7e17-40f5-b7dd-5e8c55d388a8",
                datetime.fromisoformat("2023-02-27T10:15:00+00:00"),
            ),
        ]

        histories = create_histories(repository_histories, deployment_histories)

        assert histories == [
            History(
                datetime.fromisoformat("2023-02-25T20:00:00+00:00"),
                datetime.fromisoformat("2023-02-26T22:00:00+00:00"),
                None,
            ),
        ]



import typer
from rich import print

from depwatch.command import generate_histories
from depwatch.date_utils import DateRange


app = typer.Typer()


@app.command()
def main(
    name: str,
    code_only: bool = typer.Option(
        False, help="do not retrieve deployment histories from CI"
    ),
    limit: int = typer.Option(100, help="count limit for retrieving history"),
    created_at: DateRange = typer.Option(
        None,
        parser=lambda str: DateRange.from_str(str),
        help="The range of the period when the pull requests were created",
    ),
    workflow_name: str = typer.Option(
        None,
        help="The workflow name of the CI to be obtained. This is useful when there are multiple workflows triggered by commits to the main branch.",
    ),
):
    generate_histories(name, code_only, limit, created_at, workflow_name)
    print(":sparkles::sparkles: [green]Completed![/green] :sparkles::sparkles:")


def main_cli():
    app()


if __name__ == "__main__":
    app()



from unittest.mock import patch, Mock
from depwatch.date_utils import DateRange
from depwatch.main import app
from typer.testing import CliRunner


runner = CliRunner()


class TestMain:
    @patch("depwatch.main.generate_histories")
    def test_main(self, mock_generate_histories: Mock):
        result = runner.invoke(app, ["hamakou108/my_project"])

        assert "✨✨ Completed! ✨✨" in result.stdout
        assert result.exit_code == 0
        mock_generate_histories.assert_called_once_with(
            "hamakou108/my_project", False, 100, None, None
        )

    @patch("depwatch.main.generate_histories")
    def test_main_with_all_args(self, mock_generate_histories: Mock):
        result = runner.invoke(
            app,
            [
                "hamakou108/my_project",
                "--code-only",
                "--limit",
                "10",
                "--created-at",
                "2023-01-01..2023-03-31",
                "--workflow-name",
                "deploy",
            ],
        )

        assert result.exit_code == 0

        mock_generate_histories.assert_called_once_with(
            "hamakou108/my_project",
            True,
            10,
            DateRange.from_str("2023-01-01..2023-03-31"),
            "deploy",
        )



from datetime import datetime, timezone
import secrets
from unittest.mock import patch, Mock, MagicMock
from depwatch import repository
from depwatch.date_utils import DateRange
from depwatch.exception import DepwatchException


class TestRepository:
    @patch("depwatch.repository.Github")
    def test_get_main_branch_with_main_branch(self, mock_Github: Mock):
        mock_Github.return_value.get_repo.return_value = (
            self.create_mock_repo_with_branches(["main"])
        )

        assert repository.get_main_branch("hamakou108/my_project") == "main"

    @patch("depwatch.repository.Github")
    def test_get_main_branch_with_master_branch(self, mock_Github: Mock):
        mock_Github.return_value.get_repo.return_value = (
            self.create_mock_repo_with_branches(["master"])
        )

        assert repository.get_main_branch("hamakou108/my_project") == "master"

    @patch("depwatch.repository.Github")
    def test_get_main_branch_with_other_branch(self, mock_Github: Mock):
        mock_Github.return_value.get_repo.return_value = (
            self.create_mock_repo_with_branches(["foo"])
        )

        try:
            repository.get_main_branch("hamakou108/my_project")
            assert False
        except DepwatchException as e:
            assert str(e) == "'main' or 'master' branch was not found"

    @patch("depwatch.repository.Github")
    def test_get_repository_history(self, mock_Github: Mock):
        mock_Github.return_value.search_issues.return_value = [
            self.create_mock_issue(),
            self.create_mock_issue(),
            self.create_mock_issue(),
        ]
        mock_repo = MagicMock()
        mock_repo.get_commit.return_value = self.create_mock_commit()
        mock_Github.return_value.get_repo.return_value = mock_repo

        result = repository.get_repository_history(
            "hamakou108/my_project",
            "main",
            100,
            DateRange.from_str("2023-01-01..2023-03-31"),
        )

        mock_Github.return_value.search_issues.assert_called_once_with(
            "repo:hamakou108/my_project type:pr is:merged created:2023-01-01..2023-03-31",
            "created",
            "desc",
        )

        assert len(result) == 3
        assert result[0].first_committed_at == datetime(2021, 1, 1, tzinfo=timezone.utc)
        assert result[0].merged_at == datetime(2021, 1, 3, tzinfo=timezone.utc)
        assert len(result[0].merge_commit_sha) == 40
        assert len(result[0].check_runs) == 1

    @patch("depwatch.repository.Github")
    def test_get_repository_history_when_the_limit_is_specified(
        self, mock_Github: Mock
    ):
        mock_Github.return_value.search_issues.return_value = [
            self.create_mock_issue(),
            self.create_mock_issue(),
            self.create_mock_issue(),
        ]
        mock_repo = MagicMock()
        mock_repo.get_commit.return_value = self.create_mock_commit(False)
        mock_Github.return_value.get_repo.return_value = mock_repo

        result = repository.get_repository_history("hamakou108/my_project", "main", 2)

        assert len(result) == 2

    @patch("depwatch.repository.Github")
    def test_get_repository_history_when_the_created_at_is_not_specified(
        self, mock_Github: Mock
    ):
        mock_Github.return_value.search_issues.return_value = [
            self.create_mock_issue(),
            self.create_mock_issue(),
            self.create_mock_issue(),
        ]
        mock_repo = MagicMock()
        mock_repo.get_commit.return_value = self.create_mock_commit(False)
        mock_Github.return_value.get_repo.return_value = mock_repo

        result = repository.get_repository_history("hamakou108/my_project", "main", 100)

        mock_Github.return_value.search_issues.assert_called_once_with(
            "repo:hamakou108/my_project type:pr is:merged ", "created", "desc"
        )

        assert len(result) == 3

    @patch("depwatch.repository.Github")
    def test_get_repository_history_if_a_merge_commit_with_no_check_runs_is_included(
        self, mock_Github: Mock
    ):
        mock_Github.return_value.search_issues.return_value = [
            self.create_mock_issue(),
        ]
        mock_repo = MagicMock()
        mock_repo.get_commit.return_value = self.create_mock_commit(False)
        mock_Github.return_value.get_repo.return_value = mock_repo

        result = repository.get_repository_history("hamakou108/my_project", "main", 100)

        assert len(result) == 1
        assert len(result[0].check_runs) == 0

    def create_mock_repo_with_branches(self, branches: list[str]):
        mock_repo = MagicMock()

        mock_branches = []
        for b in branches:
            mock_branch = MagicMock(spec=["name"])
            mock_branch.name = b
            mock_branches.append(mock_branch)

        mock_repo.get_branches.return_value = mock_branches

        return mock_repo

    def create_mock_commit(self, has_check_runs: bool = True):
        mock_check_run = MagicMock(spec=["raw_data"])
        mock_check_run.raw_data = {
            "name": "awesome_workflow",
            "external_id": '{"workflow-id":"10d1e2e1-cd4d-4662-a654-ff0cb02ad58d","actor-id":"20a538b5-8416-4419-b8da-a504c3835c4a"}',
            "app": {
                "slug": "circleci-checks",
            },
        }

        # Since PaginatedList from PyGitHub cannot be imported, mock it using a list instead.
        check_runs = [mock_check_run] if has_check_runs else []

        mock_commit = MagicMock(spec=["get_check_runs"])
        mock_commit.get_check_runs.return_value = check_runs

        return mock_commit

    def create_mock_pull(self):
        mock_pull = MagicMock(spec=["get_commits", "merged_at", "merge_commit_sha"])
        mock_commits = []
        for i in range(3):
            mock_commit = MagicMock(spec=["commit"])
            mock_commit.commit.author.date = datetime(
                2021, 1, i + 1, tzinfo=timezone.utc
            )
            mock_commits.append(mock_commit)
        mock_pull.get_commits.return_value = mock_commits
        mock_pull.merged_at = datetime(2021, 1, 3, tzinfo=timezone.utc)
        mock_pull.merge_commit_sha = secrets.token_hex(20)

        return mock_pull

    def create_mock_issue(self):
        mock_issue = MagicMock()
        mock_issue.as_pull_request.return_value = self.create_mock_pull()

        return mock_issue



from dotenv import load_dotenv
from depwatch.date_utils import DateRange

from depwatch.history import (
    convert_repository_history_to_workflow_ids,
    create_histories,
)
from depwatch.repository import get_main_branch, get_repository_history
from depwatch.deployment import get_deployment_history
from depwatch.writer import write_histories


def generate_histories(
    name: str,
    code_only: bool,
    limit: int,
    created_at: DateRange | None = None,
    workflow_name: str | None = None,
) -> None:
    load_dotenv()

    base = get_main_branch(name)

    repository_histories = get_repository_history(name, base, limit, created_at)

    # CircleCI
    deployment_histories = []
    if not code_only:
        workflow_ids = convert_repository_history_to_workflow_ids(
            repository_histories, workflow_name
        )
        deployment_histories = get_deployment_history(workflow_ids)

    write_histories(
        "output.csv",
        create_histories(repository_histories, deployment_histories, workflow_name),
    )



import os
from datetime import datetime
from pycircleci.api import Api
from requests import HTTPError
from requests.models import Response
from typing import cast
from depwatch.exception import DepwatchException

from depwatch.history import DeploymentHistory


def get_deployment_history(workflow_ids: list[str]) -> list[DeploymentHistory]:
    def __get_workflow(id):
        try:
            return ci.get_workflow(id)
        except HTTPError as e:
            if cast(Response, e.response).status_code == 404:
                raise DepwatchException("The workflow is not found")
            else:
                raise e

    def __is_succeeded_workflows(workflow) -> bool:
        return (
            workflow.get("status") == "success"
            and workflow.get("stopped_at") is not None
        )

    histories = []

    ci = Api(
        token=os.environ.get("CIRCLECI_ACCESS_TOKEN"), url="https://circleci.com/api"
    )

    for id in workflow_ids:
        try:
            workflow = __get_workflow(id)
        except DepwatchException:
            continue

        if not __is_succeeded_workflows(workflow):
            continue

        histories.append(
            DeploymentHistory(
                workflow.get("id"),
                datetime.fromisoformat(workflow.get("stopped_at")),
            )
        )

    return histories



from datetime import datetime, timezone
from unittest.mock import patch, Mock, MagicMock

from requests import HTTPError, Response
from depwatch import deployment

utc = timezone.utc


class TestDeployment:
    @patch("depwatch.deployment.Api")
    def test_get_deployment_history(self, mock_Api: Mock):
        mock_ci = MagicMock()
        mock_ci.get_workflow.side_effect = [
            {
                "id": "40020909-fd66-4657-b3e4-8872627507d2",
                "status": "success",
                "stopped_at": "2023-01-01T00:00:00.000Z",
            },
            {
                "id": "84abef06-1fa2-48ec-875e-382e9b7add78",
                "status": "success",
                "stopped_at": "2023-01-02T00:00:00.000Z",
            },
        ]
        mock_Api.return_value = mock_ci

        workflow_ids = [
            "40020909-fd66-4657-b3e4-8872627507d2",
            "84abef06-1fa2-48ec-875e-382e9b7add78",
        ]
        result = deployment.get_deployment_history(workflow_ids)

        assert len(result) == 2
        assert result[0].id == "40020909-fd66-4657-b3e4-8872627507d2"
        assert result[0].deployed_at == datetime(2023, 1, 1, tzinfo=timezone.utc)
        assert result[1].id == "84abef06-1fa2-48ec-875e-382e9b7add78"
        assert result[1].deployed_at == datetime(2023, 1, 2, tzinfo=timezone.utc)

    # Access to pipelines that are over six months old is not available on CircleCI
    @patch("depwatch.deployment.Api")
    def test_get_deployment_history_when_the_workflow_has_removed_then_the_item_is_not_appended(
        self, mock_Api: Mock
    ):
        mock_ci = MagicMock()
        response_404 = Response()
        response_404.status_code = 404
        mock_ci.get_workflow.side_effect = HTTPError(response=response_404)
        mock_Api.return_value = mock_ci

        workflow_ids = [
            "40020909-fd66-4657-b3e4-8872627507d2",
        ]
        result = deployment.get_deployment_history(workflow_ids)

        assert len(result) == 0

    @patch("depwatch.deployment.Api")
    def test_get_deployment_history_when_the_workflow_has_not_been_completed_then_the_item_is_not_appended(
        self, mock_Api: Mock
    ):
        mock_ci = MagicMock()
        mock_ci.get_workflow.return_value = {
            "id": "40020909-fd66-4657-b3e4-8872627507d2",
            "status": "success",
            "stopped_at": None,
        }
        mock_Api.return_value = mock_ci

        workflow_ids = [
            "40020909-fd66-4657-b3e4-8872627507d2",
        ]
        result = deployment.get_deployment_history(workflow_ids)

        assert len(result) == 0

    @patch("depwatch.deployment.Api")
    def test_get_deployment_history_when_the_workflow_is_not_successful_then_the_item_is_not_appended(
        self, mock_Api: Mock
    ):
        mock_ci = MagicMock()
        mock_ci.get_workflow.return_value = {
            "id": "40020909-fd66-4657-b3e4-8872627507d2",
            "status": "failed",
            "stopped_at": "2023-01-01T00:00:00.000Z",
        }
        mock_Api.return_value = mock_ci

        workflow_ids = [
            "40020909-fd66-4657-b3e4-8872627507d2",
        ]
        result = deployment.get_deployment_history(workflow_ids)

        assert len(result) == 0



from github import Github
import os
from datetime import datetime, timezone
from depwatch.date_utils import DateRange, convert_date_range_to_str_for_search_query
from depwatch.exception import DepwatchException
from depwatch.history import RepositoryHistory


def get_main_branch(name: str) -> str:
    gh = Github(os.environ.get("GITHUB_ACCESS_TOKEN"))

    repo = gh.get_repo(name)
    branches = repo.get_branches()

    if "main" in [b.name for b in branches]:
        return "main"
    elif "master" in [b.name for b in branches]:
        return "master"
    else:
        raise DepwatchException("'main' or 'master' branch was not found")


def get_repository_history(
    name: str, base: str, limit: int, created_at: DateRange | None = None
) -> list[RepositoryHistory]:
    histories = []

    gh = Github(os.environ.get("GITHUB_ACCESS_TOKEN"), per_page=100)
    repo = gh.get_repo(name)
    created_at_query = (
        f"created:{convert_date_range_to_str_for_search_query(created_at)}"
        if created_at is not None
        else ""
    )
    issues = gh.search_issues(
        f"repo:{name} type:pr is:merged {created_at_query}", "created", "desc"
    )[:limit]
    pulls = [i.as_pull_request() for i in issues]

    for p in pulls:
        if p.merged_at is None:
            continue

        first_committed_at: datetime = p.get_commits()[0].commit.author.date.replace(
            tzinfo=timezone.utc
        )
        merged_at: datetime = p.merged_at.replace(tzinfo=timezone.utc)
        merge_commit_sha: str = p.merge_commit_sha

        merge_commit = repo.get_commit(merge_commit_sha)
        check_runs = merge_commit.get_check_runs()

        histories.append(
            RepositoryHistory(
                first_committed_at,
                merged_at,
                merge_commit_sha,
                [item.raw_data for item in check_runs],
            )
        )

    return histories
